#!/usr/bin/env python
# Copyright (c) 2018-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for compositors creating masks."""

import unittest
from unittest import mock

import dask
import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.tests.utils import CustomScheduler


class TestHighCloudCompositor:
    """Test HighCloudCompositor."""

    def setup_method(self):
        """Create test data."""
        from pyresample.geometry import create_area_def
        area = create_area_def(area_id="test", projection={"proj": "latlong"},
                               center=(0, 45), width=3, height=3, resolution=35)
        self.dtype = np.float32
        self.data = xr.DataArray(
            da.from_array(np.array([[200, 250, 300], [200, 250, 300], [200, 250, 300]], dtype=self.dtype)),
            dims=("y", "x"), coords={"y": [0, 1, 2], "x": [0, 1, 2]},
            attrs={"area": area}
        )

    def test_high_cloud_compositor(self):
        """Test general default functionality of compositor."""
        from satpy.composites.mask import HighCloudCompositor
        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            comp = HighCloudCompositor(name="test")
            res = comp([self.data])
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        expexted_alpha = np.array([[1.0, 0.7142857, 0.0], [1.0, 0.625, 0.0], [1.0, 0.5555555, 0.0]])
        expected = np.stack([self.data, expexted_alpha])
        np.testing.assert_almost_equal(res.values, expected)

    def test_high_cloud_compositor_multiple_calls(self):
        """Test that the modified init variables are reset properly when calling the compositor multiple times."""
        from satpy.composites.mask import HighCloudCompositor
        comp = HighCloudCompositor(name="test")
        res = comp([self.data])
        res2 = comp([self.data])
        np.testing.assert_equal(res.values, res2.values)

    def test_high_cloud_compositor_dtype(self):
        """Test that the datatype is not altered by the compositor."""
        from satpy.composites.mask import HighCloudCompositor
        comp = HighCloudCompositor(name="test")
        res = comp([self.data])
        assert res.data.dtype == self.dtype

    def test_high_cloud_compositor_validity_checks(self):
        """Test that errors are raised for invalid input data and settings."""
        from satpy.composites.mask import HighCloudCompositor

        with pytest.raises(ValueError, match="Expected 2 `transition_min_limits` values, got 1"):
            _ = HighCloudCompositor("test", transition_min_limits=(210., ))

        with pytest.raises(ValueError, match="Expected 2 `latitude_min_limits` values, got 3"):
            _ = HighCloudCompositor("test", latitude_min_limits=(20., 40., 60.))

        with pytest.raises(ValueError, match="Expected `transition_max` to be of type float, "
                                             "is of type <class 'tuple'>"):
            _ = HighCloudCompositor("test", transition_max=(250., 300.))

        comp = HighCloudCompositor("test")
        with pytest.raises(ValueError, match="Expected 1 dataset, got 2"):
            _ = comp([self.data, self.data])


class TestLowCloudCompositor:
    """Test LowCloudCompositor."""

    def setup_method(self):
        """Create test data."""
        self.dtype = np.float32
        self.btd = xr.DataArray(
            da.from_array(np.array([[0.0, 1.0, 10.0], [0.0, 1.0, 10.0], [0.0, 1.0, 10.0]], dtype=self.dtype)),
            dims=("y", "x"), coords={"y": [0, 1, 2], "x": [0, 1, 2]}
        )
        self.bt_win = xr.DataArray(
            da.from_array(np.array([[250, 250, 250], [250, 250, 250], [150, 150, 150]], dtype=self.dtype)),
            dims=("y", "x"), coords={"y": [0, 1, 2], "x": [0, 1, 2]}
        )
        self.lsm = xr.DataArray(
            da.from_array(np.array([[0., 0., 0.], [1., 1., 1.], [0., 1., 0.]], dtype=self.dtype)),
            dims=("y", "x"), coords={"y": [0, 1, 2], "x": [0, 1, 2]}
        )

    def test_low_cloud_compositor(self):
        """Test general default functionality of compositor."""
        from satpy.composites.mask import LowCloudCompositor
        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            comp = LowCloudCompositor(name="test")
            res = comp([self.btd, self.bt_win, self.lsm])
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        expexted_alpha = np.array([[0.0, 0.25, 1.0], [0.0, 0.25, 1.0], [0.0, 0.0, 0.0]])
        expected = np.stack([self.btd, expexted_alpha])
        np.testing.assert_equal(res.values, expected)

    def test_low_cloud_compositor_dtype(self):
        """Test that the datatype is not altered by the compositor."""
        from satpy.composites.mask import LowCloudCompositor
        comp = LowCloudCompositor(name="test")
        res = comp([self.btd, self.bt_win, self.lsm])
        assert res.data.dtype == self.dtype

    def test_low_cloud_compositor_validity_checks(self):
        """Test that errors are raised for invalid input data and settings."""
        from satpy.composites.mask import LowCloudCompositor

        with pytest.raises(ValueError, match="Expected 2 `range_land` values, got 1"):
            _ = LowCloudCompositor("test", range_land=(2.0, ))

        with pytest.raises(ValueError, match="Expected 2 `range_water` values, got 1"):
            _ = LowCloudCompositor("test", range_water=(2.0,))

        comp = LowCloudCompositor("test")
        with pytest.raises(ValueError, match="Expected 3 datasets, got 2"):
            _ = comp([self.btd, self.lsm])


class TestMaskingCompositor:
    """Test case for the simple masking compositor."""

    @pytest.fixture
    def conditions_v1(self):
        """Masking conditions with string values."""
        return [{"method": "equal",
                 "value": "Cloud-free_land",
                 "transparency": 100},
                {"method": "equal",
                 "value": "Cloud-free_sea",
                 "transparency": 50}]

    @pytest.fixture
    def conditions_v2(self):
        """Masking conditions with numerical values."""
        return [{"method": "equal",
                 "value": 1,
                 "transparency": 100},
                {"method": "equal",
                 "value": 2,
                 "transparency": 50}]

    @pytest.fixture
    def conditions_v3(self):
        """Masking conditions with other numerical values."""
        return [{"method": "equal",
                 "value": 0,
                 "transparency": 100},
                {"method": "equal",
                 "value": 1,
                 "transparency": 0}]

    @pytest.fixture
    def test_data(self):
        """Test data to use with masking compositors."""
        return xr.DataArray(da.random.random((3, 3)), dims=["y", "x"])

    @pytest.fixture
    def test_ct_data(self):
        """Test 2D CT data array."""
        flag_meanings = ["Cloud-free_land", "Cloud-free_sea"]
        flag_values = da.array([1, 2])
        ct_data = da.array([[1, 2, 2],
                            [2, 1, 2],
                            [2, 2, 1]])
        ct_data = xr.DataArray(ct_data, dims=["y", "x"])
        ct_data.attrs["flag_meanings"] = flag_meanings
        ct_data.attrs["flag_values"] = flag_values
        return ct_data

    @pytest.fixture
    def value_3d_data(self):
        """Test 3D data array."""
        value_3d_data = da.array([[[1, 0, 0],
                                   [0, 1, 0],
                                   [0, 0, 1]]])
        value_3d_data = xr.DataArray(value_3d_data, dims=["bands", "y", "x"])
        return value_3d_data

    @pytest.fixture
    def value_3d_data_bands(self):
        """Test 3D data array."""
        value_3d_data = da.array([[[1, 0, 0],
                                   [0, 1, 0],
                                   [0, 0, 1]],
                                  [[1, 0, 0],
                                   [0, 1, 0],
                                   [0, 0, 1]],
                                  [[1, 0, 0],
                                   [0, 1, 0],
                                   [0, 0, 1]]])
        value_3d_data = xr.DataArray(value_3d_data, dims=["bands", "y", "x"])
        return value_3d_data

    @pytest.fixture
    def test_ct_data_v3(self, test_ct_data):
        """Set ct data to NaN where it originally is 1."""
        return test_ct_data.where(test_ct_data == 1)

    @pytest.fixture
    def reference_data(self, test_data, test_ct_data):
        """Get reference data to use in masking compositor tests."""
        # The data are set to NaN where ct is `1`
        return test_data.where(test_ct_data > 1)

    @pytest.fixture
    def reference_alpha(self):
        """Get reference alpha to use in masking compositor tests."""
        ref_alpha = da.array([[0, 0.5, 0.5],
                              [0.5, 0, 0.5],
                              [0.5, 0.5, 0]])
        return xr.DataArray(ref_alpha, dims=["y", "x"])

    def test_init(self):
        """Test the initializiation of compositor."""
        from satpy.composites.mask import MaskingCompositor

        # No transparency or conditions given raises ValueError
        with pytest.raises(ValueError, match="Masking conditions not defined."):
            _ = MaskingCompositor("name")

        # transparency defined
        transparency = {0: 100, 1: 50}
        conditions = [{"method": "equal", "value": 0, "transparency": 100},
                      {"method": "equal", "value": 1, "transparency": 50}]
        comp = MaskingCompositor("name", transparency=transparency.copy())
        assert not hasattr(comp, "transparency")
        # Transparency should be converted to conditions
        assert comp.conditions == conditions

        # conditions defined
        comp = MaskingCompositor("name", conditions=conditions.copy())
        assert comp.conditions == conditions

    def test_get_flag_value(self):
        """Test reading flag value from attributes based on a name."""
        from satpy.composites.mask import _get_flag_value

        flag_values = da.array([1, 2])
        mask = da.array([[1, 2, 2],
                         [2, 1, 2],
                         [2, 2, 1]])
        mask = xr.DataArray(mask, dims=["y", "x"])
        flag_meanings = ["Cloud-free_land", "Cloud-free_sea"]
        mask.attrs["flag_meanings"] = flag_meanings
        mask.attrs["flag_values"] = flag_values

        assert _get_flag_value(mask, "Cloud-free_land") == 1
        assert _get_flag_value(mask, "Cloud-free_sea") == 2

        flag_meanings_str = "Cloud-free_land Cloud-free_sea"
        mask.attrs["flag_meanings"] = flag_meanings_str
        assert _get_flag_value(mask, "Cloud-free_land") == 1
        assert _get_flag_value(mask, "Cloud-free_sea") == 2

    @pytest.mark.parametrize("mode", ["LA", "RGBA"])
    def test_call_numerical_transparency_data(
            self, conditions_v1, test_data, test_ct_data, reference_data,
            reference_alpha, mode):
        """Test call the compositor with numerical transparency data.

        Use parameterisation to test different image modes.
        """
        from satpy.composites.mask import MaskingCompositor
        from satpy.tests.utils import CustomScheduler

        # Test with numerical transparency data
        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            comp = MaskingCompositor("name", conditions=conditions_v1,
                                     mode=mode)
            res = comp([test_data, test_ct_data])
        assert res.mode == mode
        for m in mode.rstrip("A"):
            np.testing.assert_allclose(res.sel(bands=m), reference_data)
        np.testing.assert_allclose(res.sel(bands="A"), reference_alpha)

    @pytest.mark.parametrize("mode", ["LA", "RGBA"])
    def test_call_numerical_transparency_data_with_3d_mask_data(
            self, test_data, value_3d_data, conditions_v3, mode):
        """Test call the compositor with numerical transparency data.

        Use parameterisation to test different image modes.
        """
        from satpy.composites.mask import MaskingCompositor

        reference_data_v3 = test_data.where(value_3d_data[0] > 0)
        reference_alpha_v3 = xr.DataArray([[1., 0., 0.],
                                           [0., 1., 0.],
                                           [0., 0., 1.]])

        # Test with numerical transparency data using 3d test mask data which can be squeezed
        comp = MaskingCompositor("name", conditions=conditions_v3,
                                 mode=mode)
        res = comp([test_data, value_3d_data])
        assert res.mode == mode
        for m in mode.rstrip("A"):
            np.testing.assert_allclose(res.sel(bands=m), reference_data_v3)
        np.testing.assert_allclose(res.sel(bands="A"), reference_alpha_v3)

    @pytest.mark.parametrize("mode", ["LA", "RGBA"])
    def test_call_numerical_transparency_data_with_3d_mask_data_exception(
            self, test_data, value_3d_data_bands, conditions_v3, mode):
        """Test call the compositor with numerical transparency data, too many dimensions to squeeze.

        Use parameterisation to test different image modes.
        """
        from satpy.composites.mask import MaskingCompositor

        # Test with numerical transparency data using 3d test mask data which can not be squeezed
        comp = MaskingCompositor("name", conditions=conditions_v3,
                                 mode=mode)
        with pytest.raises(ValueError, match=".*Received 3 dimension\\(s\\) but expected 2.*"):
            comp([test_data, value_3d_data_bands])

    def test_call_named_fields(self, conditions_v2, test_data, test_ct_data,
                               reference_data, reference_alpha):
        """Test with named fields."""
        from satpy.composites.mask import MaskingCompositor
        from satpy.tests.utils import CustomScheduler

        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            comp = MaskingCompositor("name", conditions=conditions_v2)
            res = comp([test_data, test_ct_data])
        assert res.mode == "LA"
        np.testing.assert_allclose(res.sel(bands="L"), reference_data)
        np.testing.assert_allclose(res.sel(bands="A"), reference_alpha)

    def test_call_named_fields_string(
            self, conditions_v2, test_data, test_ct_data, reference_data,
            reference_alpha):
        """Test with named fields which are as a string in the mask attributes."""
        from satpy.composites.mask import MaskingCompositor
        from satpy.tests.utils import CustomScheduler

        flag_meanings_str = "Cloud-free_land Cloud-free_sea"
        test_ct_data.attrs["flag_meanings"] = flag_meanings_str
        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            comp = MaskingCompositor("name", conditions=conditions_v2)
            res = comp([test_data, test_ct_data])
        assert res.mode == "LA"
        np.testing.assert_allclose(res.sel(bands="L"), reference_data)
        np.testing.assert_allclose(res.sel(bands="A"), reference_alpha)

    def test_method_isnan(self, test_data,
                          test_ct_data, test_ct_data_v3):
        """Test "isnan" as method."""
        from satpy.composites.mask import MaskingCompositor
        from satpy.tests.utils import CustomScheduler

        conditions_v3 = [{"method": "isnan", "transparency": 100}]

        # The data are set to NaN where ct is NaN
        reference_data_v3 = test_data.where(test_ct_data == 1)
        reference_alpha_v3 = da.array([[1., 0., 0.],
                                       [0., 1., 0.],
                                       [0., 0., 1.]])
        reference_alpha_v3 = xr.DataArray(reference_alpha_v3, dims=["y", "x"])
        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            comp = MaskingCompositor("name", conditions=conditions_v3)
            res = comp([test_data, test_ct_data_v3])
        assert res.mode == "LA"
        np.testing.assert_allclose(res.sel(bands="L"), reference_data_v3)
        np.testing.assert_allclose(res.sel(bands="A"), reference_alpha_v3)

    def test_method_absolute_import(self, test_data, test_ct_data_v3):
        """Test "absolute_import" as method."""
        from satpy.composites.mask import MaskingCompositor
        from satpy.tests.utils import CustomScheduler

        conditions_v4 = [{"method": "absolute_import", "transparency": "satpy.resample"}]
        # This should raise AttributeError
        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            comp = MaskingCompositor("name", conditions=conditions_v4)
            with pytest.raises(AttributeError):
                comp([test_data, test_ct_data_v3])

    def test_rgb_dataset(self, conditions_v1, test_ct_data, reference_alpha):
        """Test RGB dataset."""
        from satpy.composites.mask import MaskingCompositor
        from satpy.tests.utils import CustomScheduler

        # 3D data array
        data = xr.DataArray(da.random.random((3, 3, 3)),
                            dims=["bands", "y", "x"],
                            coords={"bands": ["R", "G", "B"],
                                    "y": np.arange(3),
                                    "x": np.arange(3)})

        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            comp = MaskingCompositor("name", conditions=conditions_v1)
            res = comp([data, test_ct_data])
        assert res.mode == "RGBA"
        np.testing.assert_allclose(res.sel(bands="R"),
                                   data.sel(bands="R").where(test_ct_data > 1))
        np.testing.assert_allclose(res.sel(bands="G"),
                                   data.sel(bands="G").where(test_ct_data > 1))
        np.testing.assert_allclose(res.sel(bands="B"),
                                   data.sel(bands="B").where(test_ct_data > 1))
        np.testing.assert_allclose(res.sel(bands="A"), reference_alpha)

    def test_rgba_dataset(self, conditions_v2, test_ct_data, reference_alpha):
        """Test RGBA dataset."""
        from satpy.composites.mask import MaskingCompositor
        from satpy.tests.utils import CustomScheduler
        data = xr.DataArray(da.random.random((4, 3, 3)),
                            dims=["bands", "y", "x"],
                            coords={"bands": ["R", "G", "B", "A"],
                                    "y": np.arange(3),
                                    "x": np.arange(3)})

        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            comp = MaskingCompositor("name", conditions=conditions_v2)
            res = comp([data, test_ct_data])
        assert res.mode == "RGBA"
        np.testing.assert_allclose(res.sel(bands="R"),
                                   data.sel(bands="R").where(test_ct_data > 1))
        np.testing.assert_allclose(res.sel(bands="G"),
                                   data.sel(bands="G").where(test_ct_data > 1))
        np.testing.assert_allclose(res.sel(bands="B"),
                                   data.sel(bands="B").where(test_ct_data > 1))
        # The compositor should drop the original alpha band
        np.testing.assert_allclose(res.sel(bands="A"), reference_alpha)

    def test_incorrect_method(self, test_data, test_ct_data):
        """Test incorrect method."""
        from satpy.composites.mask import MaskingCompositor
        conditions = [{"method": "foo", "value": 0, "transparency": 100}]
        comp = MaskingCompositor("name", conditions=conditions)
        with pytest.raises(AttributeError):
            comp([test_data, test_ct_data])
        # Test with too few projectables.
        with pytest.raises(ValueError, match="Expected 2 datasets, got 1"):
            comp([test_data])

    def test_incorrect_mode(self, conditions_v1):
        """Test initiating with unsupported mode."""
        from satpy.composites.mask import MaskingCompositor

        # Incorrect mode raises ValueError
        with pytest.raises(ValueError, match="Invalid mode YCbCrA.  Supported modes: .*"):
            MaskingCompositor("name", conditions=conditions_v1,
                              mode="YCbCrA")


class TestLongitudeMaskingCompositor(unittest.TestCase):
    """Test case for the LongitudeMaskingCompositor compositor."""

    def test_masking(self):
        """Test longitude masking."""
        from satpy.composites.mask import LongitudeMaskingCompositor

        area = mock.MagicMock()
        lons = np.array([-180., -100., -50., 0., 50., 100., 180.])
        area.get_lonlats = mock.MagicMock(return_value=[lons, []])
        a = xr.DataArray(np.array([1, 2, 3, 4, 5, 6, 7]),
                         attrs={"area": area, "units": "K"})

        comp = LongitudeMaskingCompositor(name="test", lon_min=-40., lon_max=120.)
        expected = xr.DataArray(np.array([np.nan, np.nan, np.nan, 4, 5, 6, np.nan]))
        res = comp([a])
        np.testing.assert_allclose(res.data, expected.data)
        assert "units" in res.attrs
        assert res.attrs["units"] == "K"

        comp = LongitudeMaskingCompositor(name="test", lon_min=-40.)
        expected = xr.DataArray(np.array([np.nan, np.nan, np.nan, 4, 5, 6, 7]))
        res = comp([a])
        np.testing.assert_allclose(res.data, expected.data)

        comp = LongitudeMaskingCompositor(name="test", lon_max=120.)
        expected = xr.DataArray(np.array([1, 2, 3, 4, 5, 6, np.nan]))
        res = comp([a])
        np.testing.assert_allclose(res.data, expected.data)

        comp = LongitudeMaskingCompositor(name="test", lon_min=120., lon_max=-40.)
        expected = xr.DataArray(np.array([1, 2, 3, np.nan, np.nan, np.nan, 7]))
        res = comp([a])
        np.testing.assert_allclose(res.data, expected.data)


class TestFireMaskCompositor:
    """Test fire mask compositors."""

    def test_SimpleFireMaskCompositor(self):
        """Test the SimpleFireMaskCompositor class."""
        from satpy.composites.mask import SimpleFireMaskCompositor
        rows = 2
        cols = 2
        ir_105 = xr.DataArray(da.zeros((rows, cols), dtype=np.float32), dims=("y", "x"),
                              attrs={"name": "ir_105"})
        ir_105[0, 0] = 300
        ir_38 = xr.DataArray(da.zeros((rows, cols), dtype=np.float32), dims=("y", "x"),
                             attrs={"name": "ir_38"})
        ir_38[0, 0] = 400
        nir_22 = xr.DataArray(da.zeros((rows, cols), dtype=np.float32), dims=("y", "x"),
                              attrs={"name": "nir_22"})
        nir_22[0, 0] = 100
        vis_06 = xr.DataArray(da.zeros((rows, cols), dtype=np.float32), dims=("y", "x"),
                              attrs={"name": "vis_06"})
        vis_06[0, 0] = 5

        projectables = (ir_105, ir_38, nir_22, vis_06)

        with dask.config.set(scheduler=CustomScheduler(max_computes=0)):
            comp = SimpleFireMaskCompositor(
                "simple_fci_fire_mask",
                prerequisites=("ir_105", "ir_38", "nir_22", "vis_06"),
                standard_name="simple_fci_fire_mask",
                test_thresholds=[293, 20, 15, 340])
            res = comp(projectables)

        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.attrs["name"] == "simple_fci_fire_mask"
        assert res.data.dtype == bool

        assert np.array_equal(res.data.compute(),
                              np.array([[True, False], [False, False]]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Test CREFL rayleigh correction functions."""
import unittest


class TestCreflUtils(unittest.TestCase):
    """Test crefl_utils."""

    def test_get_atm_variables_abi(self):
        """Test getting atmospheric variables for ABI."""
        import numpy as np

        from satpy.modifiers._crefl_utils import _ABIAtmosphereVariables
        atm_vars = _ABIAtmosphereVariables(
            21.71342113, 77.14385758, 56.214566960,
            0.17690244, 6.123234e-17, 530.61332168, 405.,
            0.0043149700000000004, 0.0037296, 0.014107995000000002, 0.052349,
        )
        sphalb, rhoray, TtotraytH2O, tOG = atm_vars()
        assert abs(np.array(sphalb) - 0.045213532544630494) < 1e-10
        assert abs(rhoray - 2.2030281148621356) < 1e-10
        assert abs(TtotraytH2O - 0.30309880915889087) < 1e-10
        assert abs(tOG - 0.5969089524560548) < 1e-10



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.clavrx module."""
import os
from unittest import mock

import numpy as np
import pytest
import xarray as xr
from pyresample.geometry import AreaDefinition

from satpy.readers.core.loading import load_reader

ABI_FILE = "clavrx_OR_ABI-L1b-RadC-M6C01_G16_s20231021601173.level2.nc"
DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (5, 5)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)
DEFAULT_FILE_FLAGS = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                               dtype=np.byte).reshape(DEFAULT_FILE_SHAPE)
DEFAULT_FILE_FLAGS_BEYOND_FILL = DEFAULT_FILE_FLAGS
DEFAULT_FILE_FLAGS_BEYOND_FILL[-1][:-2] = [-127, -127, -128]
DEFAULT_FILE_FACTORS = np.array([2.0, 1.0], dtype=np.float32)
DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LAT_DATA = np.repeat([DEFAULT_LAT_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
DEFAULT_LON_DATA = np.linspace(5, 45, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LON_DATA = np.repeat([DEFAULT_LON_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
L1B_FILE = "clavrx_OR_ABI-L1b-RadC-M6C01_G16_s20231021601173"
ABI_FILE = f"{L1B_FILE}.level2.nc"
FILL_VALUE = -32768


def fake_test_content(filename, **kwargs):
    """Mimic reader input file content."""
    attrs = {
        "platform": "G16",
        "sensor": "ABI",
        # this is a Level 2 file that came from a L1B file
        "L1B": L1B_FILE,
    }

    longitude = xr.DataArray(DEFAULT_LON_DATA,
                             dims=("scan_lines_along_track_direction",
                                   "pixel_elements_along_scan_direction"),
                             attrs={"_FillValue": -999.,
                                    "SCALED": 0,
                                    "scale_factor": 1.,
                                    "add_offset": 0.,
                                    "standard_name": "longitude",
                                    "units": "degrees_east"
                                    })

    latitude = xr.DataArray(DEFAULT_LAT_DATA,
                            dims=("scan_lines_along_track_direction",
                                  "pixel_elements_along_scan_direction"),
                            attrs={"_FillValue": -999.,
                                   "SCALED": 0,
                                   "scale_factor": 1.,
                                   "add_offset": 0.,
                                   "standard_name": "latitude",
                                   "units": "degrees_south"
                                   })

    variable1 = xr.DataArray(DEFAULT_FILE_DATA.astype(np.int8),
                             dims=("scan_lines_along_track_direction",
                                   "pixel_elements_along_scan_direction"),
                             attrs={"_FillValue": -127,
                                    "SCALED": 0,
                                    "units": "1",
                                    })

    # data with fill values and a file_type alias
    variable2 = xr.DataArray(DEFAULT_FILE_DATA.astype(np.int16),
                             dims=("scan_lines_along_track_direction",
                                   "pixel_elements_along_scan_direction"),
                             attrs={"_FillValue": FILL_VALUE,
                                    "SCALED": 1,
                                    "scale_factor": 0.001861629,
                                    "add_offset": 59.,
                                    "units": "%",
                                    "valid_range": [-32767, 32767],
        # this is a Level 2 file that came from a L1B file
        "L1B": "clavrx_H08_20210603_1500_B01_FLDK_R",
    }
    )
    variable2 = variable2.where(variable2 % 2 != 0, FILL_VALUE)

    # category
    var_flags = xr.DataArray(DEFAULT_FILE_FLAGS.astype(np.int8),
                             dims=("scan_lines_along_track_direction",
                                   "pixel_elements_along_scan_direction"),
                             attrs={"SCALED": 0,
                                    "_FillValue": -127,
                                    "units": "1",
                                    "flag_values": [0, 1, 2, 3]})

    out_of_range_flags = xr.DataArray(DEFAULT_FILE_FLAGS_BEYOND_FILL.astype(np.int8),
                                      dims=("scan_lines_along_track_direction",
                                            "pixel_elements_along_scan_direction"),
                                      attrs={"SCALED": 0,
                                             "_FillValue": -127,
                                             "units": "1",
                                             "flag_values": [0, 1, 2, 3]})

    ds_vars = {
        "longitude": longitude,
        "latitude": latitude,
        "variable1": variable1,
        "refl_0_65um_nom": variable2,
        "var_flags": var_flags,
        "out_of_range_flags": out_of_range_flags,
    }
    ds = xr.Dataset(ds_vars, attrs=attrs)
    ds = ds.assign_coords({"latitude": latitude, "longitude": longitude})

    return ds


class TestCLAVRXReaderGeo:
    """Test CLAVR-X Reader with Geo files."""

    yaml_file = "clavrx.yaml"

    def setup_method(self):
        """Read fake data."""
        from satpy._config import config_search_paths
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))

    @pytest.mark.parametrize(
        ("filenames", "expected_loadables"),
        [([ABI_FILE], 1)]
    )
    def test_reader_creation(self, filenames, expected_loadables):
        """Test basic initialization."""
        with mock.patch("satpy.readers.clavrx.xr.open_dataset") as od:
            od.side_effect = fake_test_content
            r = load_reader(self.reader_configs)
            loadables = r.select_files_from_pathnames(filenames)
            assert len(loadables) == expected_loadables
            r.create_filehandlers(loadables)
            # make sure we have some files
            assert r.file_handlers

    @pytest.mark.parametrize(
        ("filenames", "expected_datasets"),
        [([ABI_FILE], ["variable1", "refl_0_65um_nom", "C02", "var_flags",
                       "out_of_range_flags", "longitude", "latitude"]), ]
    )
    def test_available_datasets(self, filenames, expected_datasets):
        """Test that variables are dynamically discovered."""
        from satpy.readers.core.loading import load_reader
        with mock.patch("satpy.readers.clavrx.xr.open_dataset") as od:
            od.side_effect = fake_test_content
            r = load_reader(self.reader_configs)
            loadables = r.select_files_from_pathnames(filenames)
            r.create_filehandlers(loadables)
            avails = list(r.available_dataset_names)
            for var_name in expected_datasets:
                assert var_name in avails

    @pytest.mark.parametrize(
        ("filenames", "loadable_ids"),
        [([ABI_FILE], ["variable1", "refl_0_65um_nom", "var_flags", "out_of_range_flags"]), ]
    )
    def test_load_all_new_donor(self, filenames, loadable_ids):
        """Test loading all test datasets with new donor."""
        with mock.patch("satpy.readers.clavrx.xr.open_dataset") as od:
            od.side_effect = fake_test_content
            r = load_reader(self.reader_configs)
            loadables = r.select_files_from_pathnames(filenames)
            r.create_filehandlers(loadables)

            with mock.patch("satpy.readers.clavrx.glob") as g, \
                    mock.patch("satpy.readers.clavrx.netCDF4.Dataset") as d:
                g.return_value = ["fake_donor.nc"]
                x = np.linspace(-0.1518, 0.1518, DEFAULT_FILE_SHAPE[1])
                y = np.linspace(0.1518, -0.1518, DEFAULT_FILE_SHAPE[0])
                proj = mock.Mock(
                    semi_major_axis=6378137,
                    semi_minor_axis=6356752.3142,
                    perspective_point_height=35791000,
                    longitude_of_projection_origin=140.7,
                    sweep_angle_axis="y",
                )
                d.return_value = fake_donor = mock.MagicMock(
                    variables={"goes_imager_projection": proj, "x": x, "y": y},
                )
                fake_donor.__getitem__.side_effect = lambda key: fake_donor.variables[key]

                datasets = r.load(loadable_ids + ["C02"])
                assert len(datasets) == len(loadable_ids)+1

                # should have file variable and one alias for reflectance
                assert "valid_range" not in datasets["variable1"].attrs
                assert "_FillValue" not in datasets["variable1"].attrs
                assert np.float32 == datasets["variable1"].dtype
                assert "valid_range" not in datasets["variable1"].attrs

                assert np.issubdtype(datasets["var_flags"].dtype, np.integer)
                assert datasets["var_flags"].attrs.get("flag_meanings") is not None
                assert "<flag_meanings_unknown>" == datasets["var_flags"].attrs.get("flag_meanings")
                assert np.issubdtype(datasets["out_of_range_flags"].dtype, np.integer)
                assert "valid_range" not in datasets["out_of_range_flags"].attrs

                assert isinstance(datasets["refl_0_65um_nom"].valid_range, list)
                assert np.float32 == datasets["refl_0_65um_nom"].dtype
                assert "_FillValue" not in datasets["refl_0_65um_nom"].attrs
                assert "valid_range" in datasets["refl_0_65um_nom"].attrs

                assert "refl_0_65um_nom" == datasets["C02"].file_key
                assert "_FillValue" not in datasets["C02"].attrs

                for v in datasets.values():
                    assert isinstance(v.area, AreaDefinition)
                    assert v.platform_name == "GOES-16"
                    assert v.sensor == "abi"

                    assert "calibration" not in v.attrs
                    assert "rows_per_scan" not in v.coords.get("longitude").attrs
                    assert "units" in v.attrs

    @pytest.mark.parametrize(
        ("filenames", "expected_loadables"),
        [([ABI_FILE], 1)]
    )
    def test_yaml_datasets(self, filenames, expected_loadables):
        """Test available_datasets with fake variables from YAML."""
        with mock.patch("satpy.readers.clavrx.xr.open_dataset") as od:
            od.side_effect = fake_test_content
            r = load_reader(self.reader_configs)
            loadables = r.select_files_from_pathnames(filenames)
            r.create_filehandlers(loadables)

            with mock.patch("satpy.readers.clavrx.glob") as g, \
                    mock.patch("satpy.readers.clavrx.netCDF4.Dataset") as d:
                g.return_value = ["fake_donor.nc"]
                x = np.linspace(-0.1518, 0.1518, 5)
                y = np.linspace(0.1518, -0.1518, 5)
                proj = mock.Mock(
                    semi_major_axis=6378137,
                    semi_minor_axis=6356752.3142,
                    perspective_point_height=35791000,
                    longitude_of_projection_origin=-137.2,
                    sweep_angle_axis="x",
                )
                d.return_value = fake_donor = mock.MagicMock(
                    variables={"goes_imager_projection": proj, "x": x, "y": y},
                )
                fake_donor.__getitem__.side_effect = lambda key: fake_donor.variables[key]
            # mimic the YAML file being configured for more datasets
            fake_dataset_info = [
                (None, {"name": "yaml1", "resolution": None, "file_type": ["clavrx_nc"]}),
                (True, {"name": "yaml2", "resolution": 0.5, "file_type": ["clavrx_nc"]}),
            ]
            new_ds_infos = list(r.file_handlers["clavrx_nc"][0].available_datasets(
                fake_dataset_info))
            assert len(new_ds_infos) == 10

            # we have this and can provide the resolution
            assert (new_ds_infos[0][0])
            assert new_ds_infos[0][1]["resolution"] == 2004  # hardcoded

            # we have this, but previous file handler said it knew about it
            # and it is producing the same resolution as what we have
            assert (new_ds_infos[1][0])
            assert new_ds_infos[1][1]["resolution"] == 0.5

            # we have this, but don"t want to change the resolution
            # because a previous handler said it has it
            assert (new_ds_infos[2][0])
            assert new_ds_infos[2][1]["resolution"] == 2004

    @pytest.mark.parametrize(
        ("filenames", "loadable_ids"),
        [([ABI_FILE], ["variable1", "refl_0_65um_nom", "var_flags", "out_of_range_flags"]), ]
    )
    def test_scale_data(self, filenames, loadable_ids):
        """Test that data is scaled when necessary and not scaled data are flags."""
        from satpy.readers.clavrx import _scale_data
        with mock.patch("satpy.readers.clavrx.xr.open_dataset") as od:
            od.side_effect = fake_test_content
            r = load_reader(self.reader_configs)
            loadables = r.select_files_from_pathnames(filenames)
            r.create_filehandlers(loadables)
        with mock.patch("satpy.readers.clavrx.glob") as g, \
                mock.patch("satpy.readers.clavrx.netCDF4.Dataset") as d:
            g.return_value = ["fake_donor.nc"]
            x = np.linspace(-0.1518, 0.1518, 5)
            y = np.linspace(0.1518, -0.1518, 5)
            proj = mock.Mock(
                semi_major_axis=6378137,
                semi_minor_axis=6356752.3142,
                perspective_point_height=35791000,
                longitude_of_projection_origin=-137.2,
                sweep_angle_axis="x",
            )
            d.return_value = fake_donor = mock.MagicMock(
                variables={"goes_imager_projection": proj, "x": x, "y": y},
            )
            fake_donor.__getitem__.side_effect = lambda key: fake_donor.variables[key]

            ds_scale = ["variable1", "refl_0_65um_nom"]
            ds_no_scale = ["var_flags", "out_of_range_flags"]

            with mock.patch("satpy.readers.clavrx._scale_data", wraps=_scale_data) as scale_data:
                r.load(ds_scale)
                scale_data.assert_called()

            with mock.patch("satpy.readers.clavrx._scale_data", wraps=_scale_data) as scale_data2:
                r.load(ds_no_scale)
                scale_data2.assert_not_called()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Functions and fixture to test CF code."""
import datetime
from collections import OrderedDict

import numpy as np


def get_test_attrs():
    """Create some dataset attributes for testing purpose.

    Returns:
        Attributes, encoded attributes, encoded and flattened attributes

    """
    attrs = {
        "name": "IR_108",
        "start_time": datetime.datetime(2018, 1, 1, 0),
        "end_time": datetime.datetime(2018, 1, 1, 0, 15),
        "int": 1,
        "float": 1.0,
        "none": None,  # should be dropped
        "numpy_int": np.uint8(1),
        "numpy_float": np.float32(1),
        "numpy_bool": True,
        "numpy_void": np.void(0),
        "numpy_bytes": np.bytes_("test"),
        "numpy_string": np.str_("test"),
        "list": [1, 2, np.float64(3)],
        "nested_list": ["1", ["2", [3]]],
        "bool": True,
        "array": np.array([1, 2, 3], dtype="uint8"),
        "array_bool": np.array([True, False, True]),
        "array_2d": np.array([[1, 2], [3, 4]]),
        "array_3d": np.array([[[1, 2], [3, 4]], [[1, 2], [3, 4]]]),
        "dict": {"a": 1, "b": 2},
        "nested_dict": {"l1": {"l2": {"l3": np.array([1, 2, 3], dtype="uint8")}}},
        "raw_metadata": OrderedDict([
            ("recarray", np.zeros(3, dtype=[("x", "i4"), ("y", "u1")])),
            ("flag", np.bool_(True)),
            ("dict", OrderedDict([("a", 1), ("b", np.array([1, 2, 3], dtype="uint8"))]))
            ])
     }
    encoded = {
        "name": "IR_108",
        "start_time": "2018-01-01 00:00:00",
        "end_time": "2018-01-01 00:15:00",
        "int": 1,
        "float": 1.0,
        "numpy_int": np.uint8(1),
        "numpy_float": np.float32(1),
        "numpy_bool": "true",
        "numpy_void": "[]",
        "numpy_bytes": "test",
        "numpy_string": "test",
        "list": [1, 2, np.float64(3)],
        "nested_list": '["1", ["2", [3]]]',
        "bool": "true",
        "array": np.array([1, 2, 3], dtype="uint8"),
        "array_bool": ["true", "false", "true"],
        "array_2d": "[[1, 2], [3, 4]]",
        "array_3d": "[[[1, 2], [3, 4]], [[1, 2], [3, 4]]]",
        "dict": '{"a": 1, "b": 2}',
        "nested_dict": '{"l1": {"l2": {"l3": [1, 2, 3]}}}',
        "raw_metadata": '{"recarray": [[0, 0], [0, 0], [0, 0]], '
                        '"flag": "true", "dict": {"a": 1, "b": [1, 2, 3]}}'
        }
    encoded_flat = {
        "name": "IR_108",
        "start_time": "2018-01-01 00:00:00",
        "end_time": "2018-01-01 00:15:00",
        "int": 1,
        "float": 1.0,
        "numpy_int": np.uint8(1),
        "numpy_float": np.float32(1),
        "numpy_bool": "true",
        "numpy_void": "[]",
        "numpy_bytes": "test",
        "numpy_string": "test",
        "list": [1, 2, np.float64(3)],
        "nested_list": '["1", ["2", [3]]]',
        "bool": "true",
        "array": np.array([1, 2, 3], dtype="uint8"),
        "array_bool": ["true", "false", "true"],
        "array_2d": "[[1, 2], [3, 4]]",
        "array_3d": "[[[1, 2], [3, 4]], [[1, 2], [3, 4]]]",
        "dict_a": 1,
        "dict_b": 2,
        "nested_dict_l1_l2_l3": np.array([1, 2, 3], dtype="uint8"),
        "raw_metadata_recarray": "[[0, 0], [0, 0], [0, 0]]",
        "raw_metadata_flag": "true",
        "raw_metadata_dict_a": 1,
        "raw_metadata_dict_b": np.array([1, 2, 3], dtype="uint8")
    }
    return attrs, encoded, encoded_flat



# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Utility to generate a CF-compliant DataArray."""
import logging
import warnings

from satpy.cf.attrs import preprocess_attrs
from satpy.cf.coords import add_xy_coords_attrs, set_cf_time_info

logger = logging.getLogger(__name__)


def _handle_data_array_name(original_name, numeric_name_prefix):
    if original_name[0].isdigit():
        if numeric_name_prefix:
            new_name = numeric_name_prefix + original_name
        else:
            warnings.warn(
                f"Invalid NetCDF dataset name: {original_name} starts with a digit.",
                stacklevel=5
            )
            new_name = original_name  # occurs when numeric_name_prefix = '', None or False
    else:
        new_name = original_name
    return original_name, new_name


def _preprocess_data_array_name(dataarray, numeric_name_prefix, include_orig_name):
    """Change the DataArray name by prepending numeric_name_prefix if the name is a digit."""
    original_name = None
    named_has_changed = False
    dataarray = dataarray.copy()
    if "name" in dataarray.attrs:
        original_name = dataarray.attrs.pop("name")
        original_name, new_name = _handle_data_array_name(original_name, numeric_name_prefix)
        dataarray = dataarray.rename(new_name)
        named_has_changed = original_name != new_name

    if named_has_changed and include_orig_name:
        dataarray.attrs["original_name"] = original_name
    return dataarray


def make_cf_data_array(dataarray,
                       epoch=None,
                       flatten_attrs=False,
                       exclude_attrs=None,
                       include_orig_name=True,
                       numeric_name_prefix="CHANNEL_"):
    """Make the xr.DataArray CF-compliant.

    Args:
        dataarray (xarray.DataArray): The data array to be made CF-compliant.
        epoch (str, Optional): Reference time for encoding of time coordinates.
            If None, the default reference time is defined using `from satpy.cf.coords import EPOCH`.
        flatten_attrs (bool, Optional): If True, flatten dict-type attributes. Defaults to False.
        exclude_attrs (list, Optional): List of dataset attributes to be excluded. Defaults to None.
        include_orig_name (bool, Optional): Include the original dataset name in the netcdf variable attributes.
            Defaults to True.
        numeric_name_prefix (str, Optional): Prepend dataset name with this if starting with a digit.
            Defaults to ``"CHANNEL_"``.

    Returns:
        xarray.DataArray: A CF-compliant xr.DataArray.
    """
    dataarray = _preprocess_data_array_name(dataarray=dataarray,
                                            numeric_name_prefix=numeric_name_prefix,
                                            include_orig_name=include_orig_name)
    dataarray = preprocess_attrs(data_arr=dataarray,
                                 flatten_attrs=flatten_attrs,
                                 exclude_attrs=exclude_attrs)
    dataarray = add_xy_coords_attrs(dataarray)
    if "time" in dataarray.coords:
        dataarray = set_cf_time_info(dataarray, epoch=epoch)
    return dataarray



#!/usr/bin/env python
# Copyright (c) 2018-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for compositor core functionality."""

import unittest
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyresample import AreaDefinition

from satpy.tests.utils import assert_maximum_dask_computes


class TestMatchDataArrays:
    """Test the utility method 'match_data_arrays'."""

    def _get_test_ds(self, shape=(50, 100), dims=("y", "x")):
        """Get a fake DataArray."""
        data = da.random.random(shape, chunks=25)
        area = AreaDefinition(
            "test", "test", "test",
            {"proj": "eqc", "lon_0": 0.0,
             "lat_0": 0.0},
            shape[dims.index("x")], shape[dims.index("y")],
            (-20037508.34, -10018754.17, 20037508.34, 10018754.17))
        attrs = {"area": area}
        return xr.DataArray(data, dims=dims, attrs=attrs)

    def test_single_ds(self):
        """Test a single dataset is returned unharmed."""
        from satpy.composites.core import CompositeBase
        ds1 = self._get_test_ds()
        comp = CompositeBase("test_comp")
        ret_datasets = comp.match_data_arrays((ds1,))
        assert ret_datasets[0].identical(ds1)

    def test_mult_ds_area(self):
        """Test multiple datasets successfully pass."""
        from satpy.composites.core import CompositeBase
        ds1 = self._get_test_ds()
        ds2 = self._get_test_ds()
        comp = CompositeBase("test_comp")
        ret_datasets = comp.match_data_arrays((ds1, ds2))
        assert ret_datasets[0].identical(ds1)
        assert ret_datasets[1].identical(ds2)

    def test_mult_ds_no_area(self):
        """Test that all datasets must have an area attribute."""
        from satpy.composites.core import CompositeBase
        ds1 = self._get_test_ds()
        ds2 = self._get_test_ds()
        del ds2.attrs["area"]
        comp = CompositeBase("test_comp")
        with pytest.raises(ValueError, match="Missing 'area' attribute"):
            comp.match_data_arrays((ds1, ds2))

    def test_mult_ds_diff_area(self):
        """Test that datasets with different areas fail."""
        from satpy.composites.core import CompositeBase, IncompatibleAreas
        ds1 = self._get_test_ds()
        ds2 = self._get_test_ds()
        ds2.attrs["area"] = AreaDefinition(
            "test", "test", "test",
            {"proj": "eqc", "lon_0": 0.0,
             "lat_0": 0.0},
            100, 50,
            (-30037508.34, -20018754.17, 10037508.34, 18754.17))
        comp = CompositeBase("test_comp")
        with pytest.raises(IncompatibleAreas):
            comp.match_data_arrays((ds1, ds2))

    def test_mult_ds_diff_dims(self):
        """Test that datasets with different dimensions still pass."""
        from satpy.composites.core import CompositeBase

        # x is still 50, y is still 100, even though they are in
        # different order
        ds1 = self._get_test_ds(shape=(50, 100), dims=("y", "x"))
        ds2 = self._get_test_ds(shape=(3, 100, 50), dims=("bands", "x", "y"))
        comp = CompositeBase("test_comp")
        ret_datasets = comp.match_data_arrays((ds1, ds2))
        assert ret_datasets[0].identical(ds1)
        assert ret_datasets[1].identical(ds2)

    def test_mult_ds_diff_size(self):
        """Test that datasets with different sizes fail."""
        from satpy.composites.core import CompositeBase, IncompatibleAreas

        # x is 50 in this one, 100 in ds2
        # y is 100 in this one, 50 in ds2
        ds1 = self._get_test_ds(shape=(50, 100), dims=("x", "y"))
        ds2 = self._get_test_ds(shape=(3, 50, 100), dims=("bands", "y", "x"))
        comp = CompositeBase("test_comp")
        with pytest.raises(IncompatibleAreas):
            comp.match_data_arrays((ds1, ds2))

    def test_nondimensional_coords(self):
        """Test the removal of non-dimensional coordinates when compositing."""
        from satpy.composites.core import CompositeBase
        ds = self._get_test_ds(shape=(2, 2))
        ds["acq_time"] = ("y", [0, 1])
        comp = CompositeBase("test_comp")
        ret_datasets = comp.match_data_arrays([ds, ds])
        assert "acq_time" not in ret_datasets[0].coords

    def test_almost_equal_geo_coordinates(self):
        """Test that coordinates that are almost-equal still match.

        See https://github.com/pytroll/satpy/issues/2668 for discussion.

        Various operations like cropping and resampling can cause
        geo-coordinates (y, x) to be very slightly unequal due to floating
        point precision. This test makes sure that even in those cases we
        can still generate composites from DataArrays with these coordinates.

        """
        from satpy.composites.core import CompositeBase
        from satpy.coords import add_crs_xy_coords

        comp = CompositeBase("test_comp")
        data_arr1 = self._get_test_ds(shape=(2, 2))
        data_arr1 = add_crs_xy_coords(data_arr1, data_arr1.attrs["area"])
        data_arr2 = self._get_test_ds(shape=(2, 2))
        data_arr2 = data_arr2.assign_coords(
            x=data_arr1.coords["x"] + 0.000001,
            y=data_arr1.coords["y"],
            crs=data_arr1.coords["crs"],
        )
        # data_arr2 = add_crs_xy_coords(data_arr2, data_arr2.attrs["area"])
        # data_arr2.assign_coords(x=data_arr2.coords["x"].copy() + 1.1)
        # default xarray alignment would fail and collapse one of our dims
        assert 0 in (data_arr2 - data_arr1).shape
        new_data_arr1, new_data_arr2 = comp.match_data_arrays([data_arr1, data_arr2])
        assert 0 not in new_data_arr1.shape
        assert 0 not in new_data_arr2.shape
        assert 0 not in (new_data_arr2 - new_data_arr1).shape


class TestInlineComposites(unittest.TestCase):
    """Test inline composites."""

    def test_inline_composites(self):
        """Test that inline composites are working."""
        from satpy.composites.config_loader import load_compositor_configs_for_sensors
        comps = load_compositor_configs_for_sensors(["visir"])[0]
        # Check that "fog" product has all its prerequisites defined
        keys = comps["visir"].keys()
        fog = [comps["visir"][dsid] for dsid in keys if "fog" == dsid["name"]][0]
        assert fog.attrs["prerequisites"][0]["name"] == "_fog_dep_0"
        assert fog.attrs["prerequisites"][1]["name"] == "_fog_dep_1"
        assert fog.attrs["prerequisites"][2] == 10.8

        # Check that the sub-composite dependencies use wavelengths
        # (numeric values)
        keys = comps["visir"].keys()
        fog_dep_ids = [dsid for dsid in keys if "fog_dep" in dsid["name"]]
        assert comps["visir"][fog_dep_ids[0]].attrs["prerequisites"] == [12.0, 10.8]
        assert comps["visir"][fog_dep_ids[1]].attrs["prerequisites"] == [10.8, 8.7]

        # Check the same for SEVIRI and verify channel names are used
        # in the sub-composite dependencies instead of wavelengths
        comps = load_compositor_configs_for_sensors(["seviri"])[0]
        keys = comps["seviri"].keys()
        fog_dep_ids = [dsid for dsid in keys if "fog_dep" in dsid["name"]]
        assert comps["seviri"][fog_dep_ids[0]].attrs["prerequisites"] == ["IR_120", "IR_108"]
        assert comps["seviri"][fog_dep_ids[1]].attrs["prerequisites"] == ["IR_108", "IR_087"]


class TestSingleBandCompositor(unittest.TestCase):
    """Test the single-band compositor."""

    def setUp(self):
        """Create test data."""
        from satpy.composites.core import SingleBandCompositor
        self.comp = SingleBandCompositor(name="test")

        all_valid = np.ones((2, 2))
        self.all_valid = xr.DataArray(all_valid, dims=["y", "x"])

    def test_call(self):
        """Test calling the compositor."""
        # Dataset with extra attributes
        all_valid = self.all_valid
        all_valid.attrs["sensor"] = "foo"
        attrs = {
            "foo": "bar",
            "resolution": 333,
            "units": "K",
            "sensor": {"fake_sensor1", "fake_sensor2"},
            "calibration": "BT",
            "wavelength": 10.8
        }
        self.comp.attrs["resolution"] = None
        res = self.comp([all_valid], **attrs)
        # Verify attributes
        assert res.attrs.get("sensor") == "foo"
        assert "foo" in res.attrs
        assert res.attrs.get("foo") == "bar"
        assert "units" in res.attrs
        assert "calibration" in res.attrs
        assert "modifiers" not in res.attrs
        assert res.attrs["wavelength"] == 10.8
        assert res.attrs["resolution"] == 333


class TestGenericCompositor(unittest.TestCase):
    """Test generic compositor."""

    def setUp(self):
        """Create test data."""
        from satpy.composites.core import GenericCompositor
        self.comp = GenericCompositor(name="test")
        self.comp2 = GenericCompositor(name="test2", common_channel_mask=False)

        all_valid = np.ones((1, 2, 2))
        self.all_valid = xr.DataArray(all_valid, dims=["bands", "y", "x"])
        first_invalid = np.reshape(np.array([np.nan, 1., 1., 1.]), (1, 2, 2))
        self.first_invalid = xr.DataArray(first_invalid,
                                          dims=["bands", "y", "x"])
        second_invalid = np.reshape(np.array([1., np.nan, 1., 1.]), (1, 2, 2))
        self.second_invalid = xr.DataArray(second_invalid,
                                           dims=["bands", "y", "x"])
        wrong_shape = np.reshape(np.array([1., 1., 1.]), (1, 3, 1))
        self.wrong_shape = xr.DataArray(wrong_shape, dims=["bands", "y", "x"])

    def test_masking(self):
        """Test masking in generic compositor."""
        # Single channel
        res = self.comp([self.all_valid])
        np.testing.assert_allclose(res.data, 1., atol=1e-9)
        # Three channels, one value invalid
        res = self.comp([self.all_valid, self.all_valid, self.first_invalid])
        correct = np.reshape(np.array([np.nan, 1., 1., 1.]), (2, 2))
        for i in range(3):
            np.testing.assert_almost_equal(res.data[i, :, :], correct)
        # Three channels, two values invalid
        res = self.comp([self.all_valid, self.first_invalid, self.second_invalid])
        correct = np.reshape(np.array([np.nan, np.nan, 1., 1.]), (2, 2))
        for i in range(3):
            np.testing.assert_almost_equal(res.data[i, :, :], correct)

    def test_concat_datasets(self):
        """Test concatenation of datasets."""
        from satpy.composites.core import IncompatibleAreas
        res = self.comp._concat_datasets([self.all_valid], "L")
        num_bands = len(res.bands)
        assert num_bands == 1
        assert res.shape[0] == num_bands
        assert res.bands[0] == "L"
        res = self.comp._concat_datasets([self.all_valid, self.all_valid], "LA")
        num_bands = len(res.bands)
        assert num_bands == 2
        assert res.shape[0] == num_bands
        assert res.bands[0] == "L"
        assert res.bands[1] == "A"
        with pytest.raises(IncompatibleAreas):
            self.comp._concat_datasets([self.all_valid, self.wrong_shape], "LA")

    def test_get_sensors(self):
        """Test getting sensors from the dataset attributes."""
        res = self.comp._get_sensors([self.all_valid])
        assert res is None
        dset1 = self.all_valid
        dset1.attrs["sensor"] = "foo"
        res = self.comp._get_sensors([dset1])
        assert res == "foo"
        dset2 = self.first_invalid
        dset2.attrs["sensor"] = "bar"
        res = self.comp._get_sensors([dset1, dset2])
        assert "foo" in res
        assert "bar" in res
        assert len(res) == 2
        assert isinstance(res, set)

    @mock.patch("satpy.composites.core.GenericCompositor._get_sensors")
    @mock.patch("satpy.composites.core.combine_metadata")
    @mock.patch("satpy.composites.core.check_times")
    @mock.patch("satpy.composites.core.GenericCompositor.match_data_arrays")
    def test_call_with_mock(self, match_data_arrays, check_times, combine_metadata, get_sensors):
        """Test calling generic compositor."""
        from satpy.composites.core import IncompatibleAreas
        combine_metadata.return_value = dict()
        get_sensors.return_value = "foo"
        # One dataset, no mode given
        res = self.comp([self.all_valid])
        assert res.shape[0] == 1
        assert res.attrs["mode"] == "L"
        match_data_arrays.assert_not_called()
        # This compositor has been initialized without common masking, so the
        # masking shouldn't have been called
        projectables = [self.all_valid, self.first_invalid, self.second_invalid]
        match_data_arrays.return_value = projectables
        res = self.comp2(projectables)
        match_data_arrays.assert_called_once()
        match_data_arrays.reset_mock()
        # Dataset for alpha given, so shouldn't be masked
        projectables = [self.all_valid, self.all_valid]
        match_data_arrays.return_value = projectables
        res = self.comp(projectables)
        match_data_arrays.assert_called_once()
        match_data_arrays.reset_mock()
        # When areas are incompatible, masking shouldn't happen
        match_data_arrays.side_effect = IncompatibleAreas()
        with pytest.raises(IncompatibleAreas):
            self.comp([self.all_valid, self.wrong_shape])
        match_data_arrays.assert_called_once()

    def test_call(self):
        """Test calling generic compositor."""
        # Multiple datasets with extra attributes
        all_valid = self.all_valid
        all_valid.attrs["sensor"] = "foo"
        attrs = {"foo": "bar", "resolution": 333}
        self.comp.attrs["resolution"] = None
        res = self.comp([self.all_valid, self.first_invalid], **attrs)
        # Verify attributes
        assert res.attrs.get("sensor") == "foo"
        assert "foo" in res.attrs
        assert res.attrs.get("foo") == "bar"
        assert "units" not in res.attrs
        assert "calibration" not in res.attrs
        assert "modifiers" not in res.attrs
        assert res.attrs["wavelength"] is None
        assert res.attrs["mode"] == "LA"
        assert res.attrs["resolution"] == 333

    def test_deprecation_warning(self):
        """Test deprecation warning for dcprecated composite recipes."""
        warning_message = "foo is a deprecated composite. Use composite bar instead."
        self.comp.attrs["deprecation_warning"] = warning_message
        with pytest.warns(UserWarning, match=warning_message):
            self.comp([self.all_valid])


@pytest.mark.parametrize(
    ("data_bands", "new_bands", "exp_bands"),
    [
        (["L"], ["R", "G", "B"], ["R", "G", "B"]),
        (["L"], ["R", "G", "B", "A"], ["R", "G", "B", "A"]),
        (["L", "A"], ["R", "G", "B"], ["R", "G", "B", "A"]),
        (["L", "A"], ["R", "G", "B", "A"], ["R", "G", "B", "A"]),
        (["R", "G", "B"], ["R", "G", "B", "A"], ["R", "G", "B", "A"]),
        (["R", "G", "B", "A"], ["R", "G", "B", "A"], ["R", "G", "B", "A"]),
        (["R", "G", "B", "A"], ["R", "G", "B"], ["R", "G", "B", "A"]),
    ]
)
def test_add_bands(data_bands, new_bands, exp_bands):
    """Test adding bands."""
    from satpy.composites.core import add_bands

    data_arr = xr.DataArray(
        da.ones((len(data_bands), 3, 3), dtype="float32"),
        dims=("bands", "y", "x"),
        attrs={"mode": "".join(data_bands)},
        coords={"bands": data_bands},
    )
    new_bands_arr = xr.DataArray(
        da.arange(len(new_bands)),
        dims=("bands"),
        coords={"bands": new_bands},
    )
    with assert_maximum_dask_computes(0):
        res = add_bands(data_arr, new_bands_arr.coords["bands"])
    _check_add_band_results(res, exp_bands, np.float32)


def test_add_bands_sequence():
    """Test adding bands with a sequence for the desired bands."""
    from satpy.composites.core import add_bands

    data_bands = ["L"]
    new_bands = ["R", "G", "B", "A"]
    exp_bands = ["R", "G", "B", "A"]

    data_arr = xr.DataArray(
        da.ones((len(data_bands), 3, 3), dtype="float32"),
        dims=("bands", "y", "x"),
        attrs={"mode": "".join(data_bands)},
        coords={"bands": data_bands},
    )
    with assert_maximum_dask_computes(0):
        res = add_bands(data_arr, new_bands)
    _check_add_band_results(res, exp_bands, np.float32)

def test_add_bands_p_l():
    """Test failing P to L case."""
    from satpy.composites.core import add_bands

    # P(RGBA) + L -> RGBA
    data = xr.DataArray(
        da.ones((1, 3, 3)),
        dims=("bands", "y", "x"),
        coords={"bands": ["P"]},
        attrs={"mode": "P"},
    )
    new_bands = xr.DataArray(
        da.arange(1),
        dims=("bands"),
        coords={"bands": ["L"]},
    )
    with pytest.raises(NotImplementedError), assert_maximum_dask_computes(0):
        add_bands(data, new_bands.coords["bands"])


def _check_add_band_results(res, res_bands, dtype):
    assert res.attrs["mode"] == "".join(res_bands)
    np.testing.assert_array_equal(res.bands, res_bands)
    np.testing.assert_array_equal(res.coords["bands"], res_bands)
    assert res.dtype == dtype


class TestEnhance2Dataset(unittest.TestCase):
    """Test the enhance2dataset utility."""

    @mock.patch("satpy.enhancements.enhancer.get_enhanced_image")
    def test_enhance_p_to_rgb(self, get_enhanced_image):
        """Test enhancing a paletted dataset in RGB mode."""
        from trollimage.xrimage import XRImage
        img = XRImage(xr.DataArray(np.ones((1, 20, 20)) * 2, dims=("bands", "y", "x"), coords={"bands": ["P"]}))
        img.palette = ((0, 0, 0), (4, 4, 4), (8, 8, 8))
        get_enhanced_image.return_value = img

        from satpy.composites.core import enhance2dataset
        dataset = xr.DataArray(np.ones((1, 20, 20)))
        res = enhance2dataset(dataset, convert_p=True)
        assert res.attrs["mode"] == "RGB"

    @mock.patch("satpy.enhancements.enhancer.get_enhanced_image")
    def test_enhance_p_to_rgba(self, get_enhanced_image):
        """Test enhancing a paletted dataset in RGBA mode."""
        from trollimage.xrimage import XRImage
        img = XRImage(xr.DataArray(np.ones((1, 20, 20)) * 2, dims=("bands", "y", "x"), coords={"bands": ["P"]}))
        img.palette = ((0, 0, 0, 255), (4, 4, 4, 255), (8, 8, 8, 255))
        get_enhanced_image.return_value = img

        from satpy.composites.core import enhance2dataset
        dataset = xr.DataArray(np.ones((1, 20, 20)))
        res = enhance2dataset(dataset, convert_p=True)
        assert res.attrs["mode"] == "RGBA"

    @mock.patch("satpy.enhancements.enhancer.get_enhanced_image")
    def test_enhance_p(self, get_enhanced_image):
        """Test enhancing a paletted dataset in P mode."""
        from trollimage.xrimage import XRImage
        img = XRImage(xr.DataArray(np.ones((1, 20, 20)) * 2, dims=("bands", "y", "x"), coords={"bands": ["P"]}))
        img.palette = ((0, 0, 0, 255), (4, 4, 4, 255), (8, 8, 8, 255))
        get_enhanced_image.return_value = img

        from satpy.composites.core import enhance2dataset
        dataset = xr.DataArray(np.ones((1, 20, 20)))
        res = enhance2dataset(dataset)
        assert res.attrs["mode"] == "P"
        assert res.max().values == 2

    @mock.patch("satpy.enhancements.enhancer.get_enhanced_image")
    def test_enhance_l(self, get_enhanced_image):
        """Test enhancing a paletted dataset in P mode."""
        from trollimage.xrimage import XRImage
        img = XRImage(xr.DataArray(np.ones((1, 20, 20)) * 2, dims=("bands", "y", "x"), coords={"bands": ["L"]}))
        get_enhanced_image.return_value = img

        from satpy.composites.core import enhance2dataset
        dataset = xr.DataArray(np.ones((1, 20, 20)))
        res = enhance2dataset(dataset)
        assert res.attrs["mode"] == "L"
        assert res.max().values == 1


class TestInferMode(unittest.TestCase):
    """Test the infer_mode utility."""

    def test_bands_coords_is_used(self):
        """Test that the `bands` coord is used."""
        from satpy.composites.core import GenericCompositor
        arr = xr.DataArray(np.ones((1, 5, 5)), dims=("bands", "x", "y"), coords={"bands": ["P"]})
        assert GenericCompositor.infer_mode(arr) == "P"

        arr = xr.DataArray(np.ones((3, 5, 5)), dims=("bands", "x", "y"), coords={"bands": ["Y", "Cb", "Cr"]})
        assert GenericCompositor.infer_mode(arr) == "YCbCr"

    def test_mode_is_used(self):
        """Test that the `mode` attribute is used."""
        from satpy.composites.core import GenericCompositor
        arr = xr.DataArray(np.ones((1, 5, 5)), dims=("bands", "x", "y"), attrs={"mode": "P"})
        assert GenericCompositor.infer_mode(arr) == "P"

    def test_band_size_is_used(self):
        """Test that the band size is used."""
        from satpy.composites.core import GenericCompositor
        arr = xr.DataArray(np.ones((2, 5, 5)), dims=("bands", "x", "y"))
        assert GenericCompositor.infer_mode(arr) == "LA"

    def test_no_bands_is_l(self):
        """Test that default (no band) is L."""
        from satpy.composites.core import GenericCompositor
        arr = xr.DataArray(np.ones((5, 5)), dims=("x", "y"))
        assert GenericCompositor.infer_mode(arr) == "L"



# Copyright (c) 2017-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit testing for AHI-specific enhancements functions."""

import dask.array as da
import numpy as np
import pytest
import xarray as xr


class TestTCREnhancement:
    """Test the AHI enhancement functions."""

    def setup_method(self):
        """Create test data."""
        data = da.arange(-100, 1000, 110).reshape(2, 5)
        rgb_data = np.stack([data, data, data])
        self.rgb = xr.DataArray(rgb_data, dims=("bands", "y", "x"),
                                coords={"bands": ["R", "G", "B"]},
                                attrs={"platform_name": "Himawari-8"})

    def test_jma_true_color_reproduction(self):
        """Test the jma_true_color_reproduction enhancement."""
        from trollimage.xrimage import XRImage

        from satpy.enhancements.ahi import jma_true_color_reproduction

        expected = [[[-109.93, 10.993, 131.916, 252.839, 373.762],
                     [494.685, 615.608, 736.531, 857.454, 978.377]],

                    [[-97.73, 9.773, 117.276, 224.779, 332.282],
                     [439.785, 547.288, 654.791, 762.294, 869.797]],

                    [[-93.29, 9.329, 111.948, 214.567, 317.186],
                     [419.805, 522.424, 625.043, 727.662, 830.281]]]

        img = XRImage(self.rgb)
        jma_true_color_reproduction(img)

        np.testing.assert_almost_equal(img.data.compute(), expected)

        self.rgb.attrs["platform_name"] = None
        img = XRImage(self.rgb)
        with pytest.raises(ValueError, match="Missing platform name."):
            jma_true_color_reproduction(img)

        self.rgb.attrs["platform_name"] = "Fakesat"
        img = XRImage(self.rgb)
        with pytest.raises(KeyError, match="No conversion matrix found for platform Fakesat"):
            jma_true_color_reproduction(img)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.
"""EUMETSAT EPS-SG Visible/Infrared Imager (VII) Level 1B products reader.

The ``vii_l1b_nc`` reader reads and calibrates EPS-SG VII L1b image data in netCDF format. The format is explained
in the `EPS-SG VII Level 1B Product Format Specification V4A`_.

This version is applicable for the vii test data V2 to be released in Jan 2022.

.. _EPS-SG VII Level 1B Product Format Specification V4A: https://www.eumetsat.int/media/44393

"""

import logging

import numpy as np
import xarray as xr

from satpy.readers.core.vii import C1, C2, MEAN_EARTH_RADIUS
from satpy.readers.core.vii_nc import ViiNCBaseFileHandler

logger = logging.getLogger(__name__)


class ViiL1bNCFileHandler(ViiNCBaseFileHandler):
    """Reader class for VII L1B products in netCDF format."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Read the calibration data and prepare the class for dataset reading."""
        super().__init__(filename, filename_info, filetype_info, **kwargs)

        # Read the variables which are required for the calibration
        self._bt_conversion_a = self["data/calibration_data/bt_conversion_a"].values
        self._bt_conversion_b = self["data/calibration_data/bt_conversion_b"].values
        self._channel_cw_thermal = self["data/calibration_data/channel_cw_thermal"].values
        self._integrated_solar_irradiance = self["data/calibration_data/band_averaged_solar_irradiance"].values
        # Computes the angle factor for reflectance calibration as inverse of cosine of solar zenith angle
        # (the values in the product file are on tie points and in degrees,
        # therefore interpolation and conversion to radians are required)

    def _perform_calibration(self, variable: xr.DataArray, dataset_info: dict) -> xr.DataArray:
        """Perform the calibration.

        Args:
            variable: xarray DataArray containing the dataset to calibrate.
            dataset_info: dictionary of information about the dataset.

        Returns:
            array containing the calibrated values and all the original metadata.

        """
        calibration_name = dataset_info["calibration"]
        if calibration_name == "brightness_temperature":
            # Extract the values of calibration coefficients for the current channel
            chan_index = dataset_info["chan_thermal_index"]
            cw = self._channel_cw_thermal[chan_index]
            a = self._bt_conversion_a[chan_index]
            b = self._bt_conversion_b[chan_index]
            # Perform the calibration
            calibrated_variable = self._calibrate_bt(variable, cw, a, b)
            calibrated_variable.attrs = variable.attrs
        elif calibration_name == "reflectance":
            # Extract the values of calibration coefficients for the current channel
            chan_index = dataset_info["chan_solar_index"]
            isi = self._integrated_solar_irradiance[chan_index]
            # Perform the calibration
            calibrated_variable = self._calibrate_refl(variable, isi)
            calibrated_variable.attrs = variable.attrs
        elif calibration_name == "radiance":
            calibrated_variable = variable
        else:
            raise ValueError("Unknown calibration %s for dataset %s" % (calibration_name, dataset_info["name"]))

        return calibrated_variable

    def _perform_orthorectification(self, variable: xr.DataArray, orthorect_data_name: str) -> xr.DataArray:
        """Perform the orthorectification.

        Args:
            variable: xarray DataArray containing the dataset to correct for orthorectification.
            orthorect_data_name: name of the orthorectification correction data in the product.

        Returns:
            array containing the corrected values and all the original metadata.

        """
        try:
            orthorect_data = self[orthorect_data_name]
            # Convert the orthorectification delta values from meters to degrees
            # based on the simplified formula using mean Earth radius
            variable += np.degrees(orthorect_data / MEAN_EARTH_RADIUS)
        except KeyError:
            logger.warning("Required dataset %s for orthorectification not available, skipping", orthorect_data_name)
        return variable

    @staticmethod
    def _calibrate_bt(radiance: np.ndarray, cw: float, a: float, b: float) -> np.ndarray:
        """Perform the calibration to brightness temperature.

        Args:
            radiance: numpy ndarray containing the radiance values.
            cw: center wavelength [μm].
            a: temperature coefficient [-].
            b: temperature coefficient [K].

        Returns:
            array containing the calibrated brightness temperature values.

        """
        log_expr = np.log(1.0 + C1 / ((cw ** 5) * radiance))
        bt_values = b + (a * C2 / (cw * log_expr))
        return bt_values

    @staticmethod
    def _calibrate_refl(radiance: np.ndarray, isi: float) -> np.ndarray:
        """Perform the calibration to reflectance.

        Args:
            radiance: numpy ndarray containing the radiance values.
            isi: integrated solar irradiance [W/(m2 * μm)].

        Returns:
            array containing the calibrated reflectance values.

        """
        refl_values = (np.pi / isi) * radiance * 100.0
        return refl_values



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2024, 2025 Satpy developers

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the EPS-Sterna MWR l1b filehandlers."""

from enum import Enum

import numpy as np
import pytest

from satpy.tests.reader_tests.conftest import make_fake_mwr_lonlats

geo_dims = ["n_scans", "n_fovs", "n_feedhorns"]
geo_size = 10*145*4
shape = (10, 145, 4)
fake_lon_data, fake_lat_data = make_fake_mwr_lonlats(geo_size, geo_dims, shape)


@pytest.mark.parametrize(("id_name", "file_key", "fake_array"),
                         [("longitude", "data/navigation/longitude", fake_lon_data * 1e-4),
                          ("latitude", "data/navigation/latitude", fake_lat_data),
                          ])
def test_get_navigation_data(eps_sterna_mwr_handler, id_name, file_key, fake_array):
    """Test retrieving the geolocation (lon-lat) data."""
    Horn = Enum("Horn", ["1", "2", "3", "4"])
    did = dict(name=id_name, horn=Horn["1"])
    dataset_info = dict(file_key=file_key, standard_name=id_name)
    res = eps_sterna_mwr_handler.get_dataset(did, dataset_info)
    if id_name == "longitude":
        fake_array = fake_array.where(fake_array <= 180, fake_array - 360)

    np.testing.assert_allclose(res, fake_array.isel(n_feedhorns=0))
    assert "x" in res.dims
    assert "y" in res.dims
    assert "orbital_parameters" in res.attrs
    assert res.dims == ("y", "x")
    assert "standard_name" in res.attrs
    assert "n_feedhorns" not in res.coords
    if id_name == "longitude":
        assert res.max() <= 180


def test_try_get_data_not_in_file(eps_sterna_mwr_handler):
    """Test retrieving a data field that is not available in the file."""
    did = dict(name="aws_toa_brightness_temperature")
    dataset_info = dict(file_key="data/calibration/aws_toa_brightness_temperature")

    match_str = "Dataset aws_toa_brightness_temperature not available or not supported yet!"
    with pytest.raises(NotImplementedError, match=match_str):
        _ = eps_sterna_mwr_handler.get_dataset(did, dataset_info)

def test_metadata(eps_sterna_mwr_handler):
    """Test that the metadata is read correctly."""
    assert eps_sterna_mwr_handler.sensor == "mwr"
    assert eps_sterna_mwr_handler.platform_name == "ST01"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2024-2025 Satpy developers

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for ESA Arctic Weather Satellite (AWS) level-1c file reading."""


import numpy as np
import pytest

from satpy.tests.reader_tests.conftest import make_fake_angles, make_fake_dem, make_fake_lsm, make_fake_mwr_l1c_lonlats

PLATFORM_NAME = "AWS1"

geo_dims = ["n_scans", "n_fovs"]
geo_size = 10 * 145
fake_lon_data, fake_lat_data = make_fake_mwr_l1c_lonlats(geo_size, geo_dims)
fake_sun_azi_data = make_fake_angles(geo_size, geo_dims, shape=(10, 145))
fake_sun_zen_data = make_fake_angles(geo_size, geo_dims, shape=(10, 145))
fake_sat_azi_data = make_fake_angles(geo_size, geo_dims, shape=(10, 145))
fake_sat_zen_data = make_fake_angles(geo_size, geo_dims, shape=(10, 145))

fake_lsm_data = make_fake_lsm(geo_size, geo_dims, shape=(10, 145))
fake_dem_data = make_fake_dem(geo_size, geo_dims, shape=(10, 145))


def test_get_channel_data(aws_mwr_l1c_handler, fake_mwr_data_array):
    """Test retrieving the channel data."""
    did = dict(name="1")
    dataset_info = dict(file_key="data/calibration/aws_toa_brightness_temperature")
    expected = fake_mwr_data_array.isel(n_channels=0)
    # mask no_data value
    expected = expected.where(expected != -2147483648)
    # mask outside the valid range
    expected = expected.where(expected <= 700000)
    expected = expected.where(expected >= 0)
    # "calibrate"
    expected = expected * 0.001
    res = aws_mwr_l1c_handler.get_dataset(did, dataset_info)
    np.testing.assert_allclose(res, expected)
    assert "x" in res.dims
    assert "y" in res.dims
    assert res.dims == ("y", "x")
    assert "n_channels" not in res.coords
    assert res.attrs["sensor"] == "mwr"
    assert res.attrs["platform_name"] == PLATFORM_NAME


@pytest.mark.parametrize(("id_name", "file_key", "fake_array"),
                         [("longitude", "data/navigation/aws_lon", fake_lon_data * 1e-4),
                          ("latitude", "data/navigation/aws_lat", fake_lat_data),
                          ])
def test_get_navigation_data(aws_mwr_l1c_handler, id_name, file_key, fake_array):
    """Test retrieving the geolocation (lon, lat) data."""
    did = dict(name=id_name)
    dataset_info = dict(file_key=file_key, standard_name=id_name)
    res = aws_mwr_l1c_handler.get_dataset(did, dataset_info)
    if id_name == "longitude":
        fake_array = fake_array.where(fake_array <= 180, fake_array - 360)

    np.testing.assert_allclose(res, fake_array)
    assert "x" in res.dims
    assert "y" in res.dims
    assert res.dims == ("y", "x")
    assert "standard_name" in res.attrs
    if id_name == "longitude":
        assert res.max() <= 180


@pytest.mark.parametrize(("id_name", "file_key", "fake_array"),
                         [("solar_azimuth_angle", "data/navigation/aws_solar_azimuth_angle", fake_sun_azi_data),
                          ("solar_zenith_angle", "data/navigation/aws_solar_zenith_angle", fake_sun_zen_data),
                          ("satellite_azimuth_angle", "data/navigation/aws_satellite_azimuth_angle", fake_sat_azi_data),
                          ("satellite_zenith_angle", "data/navigation/aws_satellite_zenith_angle", fake_sat_zen_data)])
def test_get_viewing_geometry_data(aws_mwr_l1c_handler, id_name, file_key, fake_array):
    """Test retrieving the angles_data."""
    dset_id = dict(name=id_name)
    dataset_info = dict(file_key=file_key, standard_name=id_name)
    res = aws_mwr_l1c_handler.get_dataset(dset_id, dataset_info)
    np.testing.assert_allclose(res, fake_array)
    assert "x" in res.dims
    assert "y" in res.dims
    assert res.dims == ("y", "x")
    assert "standard_name" in res.attrs

@pytest.mark.parametrize(("id_name", "file_key", "fake_array"),
                         [("surface_type", "data/navigation/aws_surface_type", fake_lsm_data),
                          ("terrain_elevation", "data/navigation/aws_terrain_elevation", fake_dem_data)])
def test_get_lsm_or_dem_data(aws_mwr_l1c_handler, id_name, file_key, fake_array):
    """Test retrieving the lsm and dem data."""
    dset_id = dict(name=id_name)
    dataset_info = dict(file_key=file_key, standard_name=id_name)
    res = aws_mwr_l1c_handler.get_dataset(dset_id, dataset_info)
    np.testing.assert_allclose(res, fake_array)
    assert "x" in res.dims
    assert "y" in res.dims
    assert res.dims == ("y", "x")
    assert "standard_name" in res.attrs

@pytest.mark.parametrize(("id_name", "file_key", "fake_array"),
                         [("unknwon_dataset", "data/navigation/not_implemented_yet", fake_dem_data)])
def test_dataset_name_not_implemented(aws_mwr_l1c_handler, id_name, file_key, fake_array):
    """Test retrieving a dataset for which reading is yet not implemented."""
    dset_id = dict(name=id_name)
    dataset_info = dict(file_key=file_key, standard_name=id_name)
    with pytest.raises(NotImplementedError, match="Dataset unknwon_dataset not available or not supported yet!"):
        _ = aws_mwr_l1c_handler.get_dataset(dset_id, dataset_info)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2023- Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""The viirs_vgac_l1b_nc reader tests package.

This version tests the readers for VIIIRS VGAC data preliminary version.

"""


import datetime as dt

import numpy as np
import pytest
import xarray as xr
from netCDF4 import Dataset


@pytest.fixture
def nc_filename(tmp_path):
    """Create an nc test data file and return its filename."""
    now = dt.datetime.now(dt.timezone.utc)
    filename = f"VGAC_VJ10XMOD_A{now:%Y%j_%H%M}_n004946_K005.nc"
    filename_str = str(tmp_path / filename)
    # Create test data
    with Dataset(filename_str, "w") as nc:
        nscn = 7
        npix = 800
        n_lut = 12000
        start_time_srting = "2023-03-28T09:08:07"
        end_time_string = "2023-03-28T10:11:12"
        nc.createDimension("npix", npix)
        nc.createDimension("nscn", nscn)
        nc.createDimension("n_lut", n_lut)
        nc.createDimension("one", 1)
        nc.StartTime = start_time_srting
        nc.EndTime = end_time_string
        for ind in range(1, 11, 1):
            ch_name = "M{:02d}".format(ind)
            r_a = nc.createVariable(ch_name, np.int16, dimensions=("nscn", "npix"))
            r_a[:] = np.ones((nscn, npix)) * 10
            attrs = {"scale_factor": 0.1, "units": "percent"}
            for attr in attrs:
                setattr(r_a, attr, attrs[attr])
        for ind in range(12, 17, 1):
            ch_name = "M{:02d}".format(ind)
            tb_b = nc.createVariable(ch_name, np.int16, dimensions=("nscn", "npix"))
            tb_b[:] = np.ones((nscn, npix)) * 800
            attrs = {"units": "radiances", "scale_factor": 0.002}
            for attr in attrs:
                setattr(tb_b, attr, attrs[attr])
            tb_lut = nc.createVariable(ch_name + "_LUT", np.float32, dimensions=("n_lut"))
            tb_lut[:] = np.array(range(0, n_lut)) * 0.5
            tb_lut.units = "Kelvin"
        reference_time = np.datetime64("2010-01-01T00:00:00")
        start_time = np.datetime64("2023-03-28T09:08:07") + np.timedelta64(123000, "us")
        delta_days = start_time - reference_time
        delta_full_days = delta_days.astype("timedelta64[D]")
        hidden_reference_time = reference_time + delta_full_days
        delta_part_of_days = start_time - hidden_reference_time
        proj_time0 = nc.createVariable("proj_time0", np.float64)
        proj_time0[:] = (delta_full_days.astype(np.int64) +
                         0.000001 * delta_part_of_days.astype("timedelta64[us]").astype(np.int64) / (60 * 60 * 24))
        proj_time0.units = "days since 01/01/2010T00:00:00"
        time_v = nc.createVariable("time", np.float64, ("nscn",))
        delta_h = np.datetime64(end_time_string) - start_time
        delta_hours = 0.000001 * delta_h.astype("timedelta64[us]").astype(np.int64) / (60 * 60)
        time_v[:] = np.linspace(0, delta_hours, num=nscn).astype(np.float64)
        time_v.units = "hours since proj_time0"

    return filename_str


class TestVGACREader:
    """Test the VGACFileHandler reader."""

    def test_read_vgac(self, nc_filename):
        """Test reading reflectances and BT."""
        from satpy.scene import Scene

        # Read data
        scn_ = Scene(
            reader="viirs_vgac_l1c_nc",
            filenames=[nc_filename])
        scn_.load(["M05", "M15", "scanline_timestamps"])
        diff_s = (scn_["scanline_timestamps"][0].values.astype("datetime64[us]") -
                  np.datetime64("2023-03-28T09:08:07.123000").astype("datetime64[us]"))
        diff_e = (np.datetime64("2023-03-28T10:11:12.000000").astype("datetime64[us]") -
                  scn_["scanline_timestamps"][-1].values.astype("datetime64[us]"))
        assert (diff_s < np.timedelta64(5, "us"))
        assert (diff_s > np.timedelta64(-5, "us"))
        assert (diff_e < np.timedelta64(5, "us"))
        assert (diff_e > np.timedelta64(-5, "us"))
        assert (scn_["M05"][0, 0] == 100)
        assert (scn_["M15"][0, 0] == 400)
        assert scn_.start_time == dt.datetime(year=2023, month=3, day=28,
                                              hour=9, minute=8, second=7)
        assert scn_.end_time == dt.datetime(year=2023, month=3, day=28,
                                            hour=10, minute=11, second=12)

    def test_decode_time_variable(self):
        """Test decode time variable branch."""
        from satpy.readers.viirs_vgac_l1c_nc import VGACFileHandler
        fh = VGACFileHandler(filename="",
                             filename_info={"start_time": "2023-03-28T09:08:07"},
                             filetype_info="")
        data = xr.DataArray(
            [[1, 2],
             [3, 4]],
            dims=("y", "x"),
            attrs={"units": "something not expected"})
        with pytest.raises(AttributeError):
            fh.decode_time_variable(data, "time", None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""ACSPO SST Reader.

See the following page for more information:

https://podaac.jpl.nasa.gov/dataset/VIIRS_NPP-OSPO-L2P-v2.3

Cloud clearing and data filtering
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Cloud clearing can be enabled for supported variables (ex. "sst") by passing
the filter name "cloud_clear" to the file handler "filters" keyword argument::

   scn = Scene(reader="ascpo", filenames=[...],
               reader_kwargs={"filters": ["cloud_clear"]})

.. versionchanged: 0.58.0

   The cloud clearing is no longer enabled by default.

"""

import datetime as dt
import logging

import numpy as np

from satpy.readers.core.netcdf import NetCDF4FileHandler

LOG = logging.getLogger(__name__)


ROWS_PER_SCAN = {
    "modis": 10,
    "viirs": 16,
    "avhrr": None,
}


class ACSPOFileHandler(NetCDF4FileHandler):
    """ACSPO L2P SST File Reader."""

    def __init__(self, filename, filename_info, filetype_info,
                 filters: list[str] | None = None, **kwargs):
        """Initialize file handler and store cloud clear flag."""
        super().__init__(filename, filename_info, filetype_info, **kwargs)
        filters = filters or []
        for filter_name in filters:
            if filter_name not in ("cloud_clear",):
                raise ValueError(f"Unknown filter '{filter_name}'")
        self.cloud_clear = "cloud_clear" in filters

    @property
    def platform_name(self):
        """Get satellite name for this file's data."""
        res = self["/attr/platform"]
        if isinstance(res, np.ndarray):
            return str(res.astype(str))
        return res

    @property
    def sensor_name(self):
        """Get instrument name for this file's data."""
        res = self["/attr/sensor"]
        if isinstance(res, np.ndarray):
            res = str(res.astype(str))
        return res.lower()

    def get_shape(self, ds_id, ds_info):
        """Get numpy array shape for the specified dataset.

        Args:
            ds_id (DataID): ID of dataset that will be loaded
            ds_info (dict): Dictionary of dataset information from config file

        Returns:
            tuple: (rows, cols)

        """
        var_path = ds_info.get("file_key", "{}".format(ds_id["name"]))
        if var_path + "/shape" not in self:
            # loading a scalar value
            shape = 1
        else:
            shape = self[var_path + "/shape"]
            if len(shape) == 3:
                if shape[0] != 1:
                    raise ValueError("Not sure how to load 3D Dataset with more than 1 time")
                shape = shape[1:]
        return shape

    @staticmethod
    def _parse_datetime(datestr):
        return dt.datetime.strptime(datestr, "%Y%m%dT%H%M%SZ")

    @property
    def start_time(self):
        """Get first observation time of data."""
        return self._parse_datetime(self["/attr/time_coverage_start"])

    @property
    def end_time(self):
        """Get final observation time of data."""
        return self._parse_datetime(self["/attr/time_coverage_end"])

    def get_metadata(self, dataset_id, ds_info):
        """Collect various metadata about the specified dataset."""
        var_path = ds_info.get("file_key", "{}".format(dataset_id["name"]))
        shape = self.get_shape(dataset_id, ds_info)
        units = self[var_path + "/attr/units"]
        info = getattr(self[var_path], "attrs", {})
        standard_name = self[var_path + "/attr/standard_name"]
        resolution = float(self["/attr/spatial_resolution"].split(" ")[0])
        rows_per_scan = ROWS_PER_SCAN.get(self.sensor_name) or 0
        info.update(dataset_id.to_dict())
        info.update({
            "shape": shape,
            "units": units,
            "platform_name": self.platform_name,
            "sensor": self.sensor_name,
            "standard_name": standard_name,
            "resolution": resolution,
            "rows_per_scan": rows_per_scan,
            "long_name": self.get(var_path + "/attr/long_name"),
            "comment": self.get(var_path + "/attr/comment"),
        })
        return info

    def get_dataset(self, dataset_id, ds_info):
        """Load data array and metadata from file on disk."""
        var_path = ds_info.get("file_key", "{}".format(dataset_id["name"]))
        metadata = self.get_metadata(dataset_id, ds_info)
        shape = metadata["shape"]
        file_shape = self[var_path + "/shape"]
        metadata["shape"] = shape

        valid_min = self[var_path + "/attr/valid_min"]
        valid_max = self[var_path + "/attr/valid_max"]
        # no need to check fill value since we are using valid min/max
        scale_factor = self.get(var_path + "/attr/scale_factor")
        add_offset = self.get(var_path + "/attr/add_offset")

        data = self[var_path]
        data = data.rename({"ni": "x", "nj": "y"})
        if isinstance(file_shape, tuple) and len(file_shape) == 3:
            # can only read 3D arrays with size 1 in the first dimension
            data = data[0]
        data = data.where((data >= valid_min) & (data <= valid_max))
        if scale_factor is not None:
            data = data * scale_factor + add_offset

        if self.cloud_clear and ds_info.get("cloud_clear", False):
            LOG.info(f"Cloud clearing {dataset_id}")
            # clear-sky if bit 15-16 are 00
            l2p_flags = self._get_unsigned_l2p_flags()
            clear_sky_mask = (l2p_flags & 0b1100000000000000) != 0
            clear_sky_mask = clear_sky_mask.rename({"ni": "x", "nj": "y"})
            data = data.where(~clear_sky_mask)

        data.attrs.update(metadata)
        # Remove these attributes since they are no longer valid and can cause invalid value filling.
        data.attrs.pop("_FillValue", None)
        data.attrs.pop("valid_max", None)
        data.attrs.pop("valid_min", None)
        return data

    def _get_unsigned_l2p_flags(self):
        l2p_flags = self["l2p_flags"][0]
        # l2p_flags is usually signed 16-bit (int16) but we need (uint16) for binary operations
        unsigned_type = l2p_flags.dtype.str.replace("i", "u")
        return l2p_flags.astype(unsigned_type, copy=False)



"""File handler for Insat 3D L1B data in hdf5 format."""

import datetime as dt
from contextlib import suppress
from functools import cached_property

import dask.array as da
import numpy as np
import xarray as xr
from xarray.core.datatree import DataTree

from satpy.readers.core.file_handlers import BaseFileHandler

LUT_SUFFIXES = {"vis": ("RADIANCE", "ALBEDO"),
                "swir": ("RADIANCE",),
                "mir": ("RADIANCE", "TEMP"),
                "tir1": ("RADIANCE", "TEMP"),
                "tir2": ("RADIANCE", "TEMP"),
                "wv": ("RADIANCE", "TEMP"),
                }

CHANNELS_BY_RESOLUTION = {1000: ["vis", "swir"],
                          4000: ["mir", "tir1", "tir2"],
                          8000: ["wv"],
                          }


def apply_lut(data, lut):
    """Apply a lookup table."""
    return lut[data]


def decode_lut_arr(arr, lut):
    """Decode an array using a lookup table."""
    dtype = lut.dtype
    lut_attrs = lut.attrs

    attrs = arr.attrs
    attrs["units"] = lut_attrs["units"]
    attrs["long_name"] = lut_attrs["long_name"]
    new_darr = da.map_blocks(apply_lut, arr.data, lut=np.asanyarray(lut), dtype=dtype)
    new_arr = xr.DataArray(new_darr, dims=arr.dims, attrs=attrs, coords=arr.coords)
    new_arr = new_arr.where(arr.data != attrs["_FillValue"])
    return new_arr


def get_lonlat_suffix(resolution):
    """Get the lonlat variable suffix from the resolution."""
    if resolution == 1000:
        lonlat_suffix = "_VIS"
    elif resolution == 8000:
        lonlat_suffix = "_WV"
    else:
        lonlat_suffix = ""
    return lonlat_suffix


def open_dataset(filename, resolution=1000):
    """Open a dataset for a given resolution."""
    if resolution not in [1000, 4000, 8000]:
        raise ValueError(f"Resolution {resolution} not available. Available resolutions: 1000, 4000, 8000")

    h5ds = xr.open_dataset(filename, engine="h5netcdf", chunks="auto")
    h5ds_raw = xr.open_dataset(filename, engine="h5netcdf", chunks="auto", mask_and_scale=False)
    ds = xr.Dataset()
    ds.attrs = h5ds.attrs
    for channel in CHANNELS_BY_RESOLUTION[resolution]:
        var_name = "IMG_" + channel.upper()
        channel_data = h5ds_raw[var_name]
        ds[var_name] = channel_data

        for name in [var_name + "_" + suffix for suffix in LUT_SUFFIXES[channel]]:
            lut = h5ds[name]
            decoded = decode_lut_arr(channel_data, lut)
            ds[name] = decoded

        lonlat_suffix = get_lonlat_suffix(resolution)

        for coord in ["Longitude", "Latitude"]:
            var_name = coord + lonlat_suffix
            ds[var_name] = h5ds[var_name]

    ds = _rename_dims(ds)
    return ds


def _rename_dims(ds):
    """Rename dimensions to satpy standards."""
    for x_dim in ["GeoX", "GeoX1", "GeoX2"]:
        with suppress(ValueError):
            ds = ds.rename({x_dim: "x"})
    for y_dim in ["GeoY", "GeoY1", "GeoY2"]:
        with suppress(ValueError):
            ds = ds.rename({y_dim: "y"})
    for lons in ["Longitude_VIS", "Longitude_WV"]:
        with suppress(ValueError):
            ds = ds.rename({lons: "Longitude"})
    for lats in ["Latitude_VIS", "Latitude_WV"]:
        with suppress(ValueError):
            ds = ds.rename({lats: "Latitude"})
    return ds


def open_datatree(filename):
    """Open a datatree."""
    datasets = {}
    for resolution in [1000, 4000, 8000]:
        datasets[str(resolution)] = open_dataset(filename, resolution)
    dt = DataTree.from_dict(datasets)
    dt.attrs = dt["1000"].attrs
    return dt


class Insat3DIMGL1BH5FileHandler(BaseFileHandler):
    """File handler for insat 3d imager data."""

    @property
    def start_time(self):
        """Get the start time."""
        start_time = dt.datetime.strptime(
            self.datatree.attrs["Acquisition_Start_Time"], "%d-%b-%YT%H:%M:%S")
        return start_time

    @property
    def end_time(self):
        """Get the end time."""
        end_time = dt.datetime.strptime(
            self.datatree.attrs["Acquisition_End_Time"], "%d-%b-%YT%H:%M:%S")
        return end_time

    @cached_property
    def datatree(self):
        """Create the datatree."""
        return open_datatree(self.filename)

    def get_dataset(self, ds_id, ds_info):
        """Get a data array."""
        resolution = ds_id["resolution"]
        ds = self.datatree[str(resolution)]
        if ds_id["name"] in ["longitude", "latitude"]:
            darr = ds[ds_id["name"].capitalize()]

            return darr

        if ds_id["calibration"] == "counts":
            calibration = ""
        elif ds_id["calibration"] == "radiance":
            calibration = "_RADIANCE"
        elif ds_id["calibration"] == "reflectance":
            calibration = "_ALBEDO"
        elif ds_id["calibration"] == "brightness_temperature":
            calibration = "_TEMP"

        darr = ds["IMG_" + ds_id["name"] + calibration]

        nlat, nlon = ds.attrs["Nominal_Central_Point_Coordinates(degrees)_Latitude_Longitude"]
        darr.attrs["orbital_parameters"] = dict(satellite_nominal_longitude=float(nlon),
                                                satellite_nominal_latitude=float(nlat),
                                                satellite_nominal_altitude=float(ds.attrs["Nominal_Altitude(km)"]),
                                                satellite_actual_altitude=float(ds.attrs["Observed_Altitude(km)"]))
        darr.attrs["platform_name"] = "insat-3d"
        darr.attrs["sensor"] = "imager"
        darr = darr.squeeze()

        return darr

    def get_area_def(self, ds_id):
        """Get the area definition."""
        from satpy.readers.core._geos_area import get_area_definition, get_area_extent
        darr = self.get_dataset(ds_id, None)
        shape = darr.shape
        lines = shape[-2]
        cols = shape[-1]

        # From empirical analysis, hardcoding the view of view to 18 degrees
        # produces better geolocation results.
        # Uncommenting the line below will use the fov from the file instead,
        # this line is kept for reference.
        #fov = self.datatree.attrs["Field_of_View(degrees)"]
        fov = 18
        cfac = 2 ** 16 / (fov / cols)

        # From reverse engineering metadata from a netcdf file, we discovered
        # the lfac is actually the same as cfac, ie dependent on cols, not lines!
        lfac = 2 ** 16 / (fov / cols)

        h = self.datatree.attrs["Observed_Altitude(km)"] * 1000
        # WGS 84
        a = 6378137.0
        b = 6356752.314245

        subsatellite_longitude = self.datatree.attrs["Nominal_Central_Point_Coordinates(degrees)_Latitude_Longitude"][1]

        pdict = {
            "cfac": cfac,
            "lfac": lfac,
            "coff": cols // 2 + 1,
            "loff": lines // 2,
            "ncols": cols,
            "nlines": lines,
            "scandir": "N2S",
            "a": a,
            "b": b,
            "h": h,
            "ssp_lon": subsatellite_longitude,
            "a_name": "insat3d82",
            "a_desc": "insat3d82",
            "p_id": "geosmsg"
        }
        area_extent = get_area_extent(pdict)
        adef = get_area_definition(pdict, area_extent)
        return adef



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for MODIS L2 HDF reader."""

from __future__ import annotations

import dask.array as da

from satpy import Scene, available_readers

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - modis_l3_nasa_mcd12q1_file


class TestModisL3MCD12Q1:
    """Test MODIS L3 MCD12Q1 reader."""

    def test_available_reader(self):
        """Test that MODIS L3 reader is available."""
        assert "mcd12q1" in available_readers()

    def test_metadata(self, modis_l3_nasa_mcd12q1_file):
        """Test some basic metadata that should exist in the file."""
        scene = Scene(reader="mcd12q1", filenames=modis_l3_nasa_mcd12q1_file)
        ds_name = "LC_Type2"
        scene.load([ds_name])
        assert scene[ds_name].attrs["area"].description == "Tiled sinusoidal L3 MODIS area"
        assert scene[ds_name].attrs["sensor"] == "modis"

    def test_scene_available_datasets(self, modis_l3_nasa_mcd12q1_file):
        """Test that datasets are available."""
        scene = Scene(reader="mcd12q1", filenames=modis_l3_nasa_mcd12q1_file)
        available_datasets = scene.all_dataset_names()
        assert len(available_datasets) > 0
        assert "LC_Type1" in available_datasets

    def test_load_l3_dataset(self, modis_l3_nasa_mcd12q1_file):
        """Load and check an L2 variable."""
        scene = Scene(reader="mcd12q1", filenames=modis_l3_nasa_mcd12q1_file)
        ds_name = "LC_Type1"
        scene.load([ds_name])
        assert ds_name in scene
        data_arr = scene[ds_name]
        assert isinstance(data_arr.data, da.Array)
        assert data_arr.attrs.get("resolution") == 500
